import { ActionFormData } from "@minecraft/server-ui";
import { system, world, ItemStack, BlockTypes } from "@minecraft/server";
import { getMagicData } from "./magic_list";
import * as ItemData from "./item_list";
import { Vector } from "./Vector3";
import { getSpellRecipes } from "./spell_recipes";

const GradeToInt = {
	basic: 0,
	intermediate: 1,
	advance: 2,
	expert: 3,
	master: 4
}

const ElementItem = {
	"amw:nature_magic_essence": "nature",
	"amw:fire_magic_essence": "fire", 
	"amw:void_magic_essence": "void", 
    "amw:dark_magic_essence": "dark",
    "amw:water_magic_essence": "water",
    "amw:thunder_magic_essence": "thunder",
    "amw:wind_magic_essence": "wind",
    "amw:light_magic_essence": "light",
    "amw:ice_magic_essence": "ice",
    "minecraft:paper": "paper"
}

const ElementToItem = {
	nature: "amw:nature_magic_essence",
    fire: "amw:fire_magic_essence", 
    void: "amw:void_magic_essence", 
    dark: "amw:dark_magic_essence",
    water: "amw:water_magic_essence",
    thunder: "amw:thunder_magic_essence",
    wind: "amw:wind_magic_essence",
    light: "amw:light_magic_essence",
    ice: "amw:ice_magic_essence"
}

const EssenseItemId = Object.keys(ElementItem);
EssenseItemId.push("minecraft:paper");

const Dimensions = [
    world.getDimension("minecraft:overworld"),
    world.getDimension("minecraft:nether"),
    world.getDimension("minecraft:the_end")
]

const grades = [
	"Unlearned",
	"Basic",
	"Intermediate",
	"Advance",
	"Expert",
	"Master"
]

var SortedSpellList = {};
var attributes = [];
var ScrollItem = {};

function getRawName(name){
	if(BlockTypes.get(name)){
		if(name.startsWith("minecraft:")){
			return "tile." + name.substr(10) + ".name";
		}else{
			return "tile." + name + ".name";
		}
	}else{
		if(name.startsWith("minecraft:")){
			return "item." + name.substr(10) + ".name";
		}else{
			return "item." + name;
		}
	}
}

export function updateMagicLearningStation(){
    let magic_data;
    try{
        magic_data = getMagicData();
    }catch(err){
		system.runTimeout(()=>{
			updateMagicLearningStation()
		}, 2);
        return;
    }
	SortedSpellList = {};
	attributes = [];
	ScrollItem = {};
	for(let list of Object.keys(magic_data)){
		let magic = getMagicData()[list];
		if(magic.attribute == undefined) continue;
	
		for(let attribute of magic.attribute){
			if(SortedSpellList[attribute] == undefined){
				SortedSpellList[attribute] = [];
				attributes.push(attribute);
			}
	
			SortedSpellList[attribute].push(list);
		}
	}
	
	for(let list of attributes){
		SortedSpellList[list].sort((a, b) => {
			let from = GradeToInt[getMagicData()[a].grade];
			let to = GradeToInt[getMagicData()[b].grade];
			return from - to;
		});
	}
	
	for(let scroll of Object.keys(ItemData.getScrolls())){
		ScrollItem[ItemData.getScrolls()[scroll]] = scroll;
	}
}
updateMagicLearningStation();

const elements_type = [
	[ "fire_type", "Fire Type", "fire", "fire_mastering" ],
	[ "dark_type", "Dark Type", "dark", "dark_mastering" ],
	[ "nature_type", "Nature Type", "nature", "nature_mastering" ],
	[ "wind_type", "Wind Type", "wind", "wind_mastering" ],
	[ "void_type", "Void Type", "void", "void_mastering" ],
	[ "ice_type", "Ice Type", "ice", "ice_mastering" ],
	[ "water_type", "Water Type", "water", "water_mastering" ],
	[ "thunder_type", "Thunder Type", "thunder", "thunder_mastering" ],
	[ "light_type", "Light Type", "light", "light_mastering" ]
];

function capitalize(string) {
    return string.charAt(0).toUpperCase() + string.slice(1);
}

function scrollCraft(source, selected, book, list_select = -1){
	let recipes = getSpellRecipes();
	let table = new ActionFormData();
    table.title('magic_learning_station.menu');
	let visibility = "menu.invisible";
	let spell_mastering = book;

	let spell_offset = 0;
	let element_id = 0;
	for(let elements of elements_type){
		if(element_id == selected){
			table.button(elements[1], elements[0] + " locked");
		}else{
			table.button(elements[1], elements[0]);
		}
		element_id++;
	}
	
	let spell_recipes;
	let item_identifier = [];
	let item_identifier_index = {};
	let can_craft = false;
	if(list_select != -1){
		
		//table.button("description_scroll", "[  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ][  Work In Progress ]");
		let spell = getMagicData()[SortedSpellList[elements_type[selected][2]][list_select]];
		let desc = "";
		spell_recipes = recipes[SortedSpellList[elements_type[selected][2]][list_select]];
		let container = source.getComponent("minecraft:inventory").container;

		let items = {};
		let i = 0;
		for(let recipe_item of spell_recipes){
			item_identifier.push(recipe_item.identifier);
			item_identifier_index[recipe_item.identifier] = i;
			items[recipe_item.identifier] = 0;
			i++;
		}
		if(source.getGameMode() == "creative"){
			item_identifier.forEach(name =>{
				items[name] = 99;
			})
		}else{
			for(let i = 0; i < 36; i++){
				let item = container.getItem(i);
				if(item != undefined && item_identifier.includes(item.typeId)){
					items[item.typeId] += item.amount;
				}
			}	
		}
		
		can_craft = true;

		let recipes_list = {
			rawtext: [
				{
					text: "Recipes: \n"
				}
			]
		};
		
		for(let recipe_item of spell_recipes){
			let text = {
				rawtext: [
					{
						text: ((items[recipe_item.identifier] < recipe_item.amount || items[recipe_item.identifier] == undefined) ? "§c" : "§r")
					},
					{
						translate: getRawName(recipe_item.identifier)
					},
					{
						text: " x" + recipe_item.amount + "\n"
					}
				]
			};
			
			if(items[recipe_item.identifier] < recipe_item.amount || items[recipe_item.identifier] == undefined) can_craft = false;
			recipes_list.rawtext.push(text);
		}
		spell_offset += 2;
		table.button(recipes_list, "scroll_recipe");
		desc += "[ " + spell.name + " ]\n";
		desc += "Cast Duration: " + (spell.cast_duration / 20).toFixed(1) + "s\n";
		desc += "Soul Cost: " + spell.soul_cost + "\n\n";
		desc += "Attributes:\n";

		let gap = false;
		for(let attribute of spell.attribute){
			if(gap) desc += ", ";
			desc += capitalize(attribute);
			if(gap == false) gap = true;
		}
		desc += "\nType: " + capitalize(spell.type) + "\n\n";
		desc += "Description:\n";
		desc += spell.description;
		table.button(desc + "\n\n", "description_scroll");
		if(can_craft){
			table.button("craft", "craft");
			spell_offset += 1;
		}
	}

	let locked_id = ["", ""];
	let spell_id = 0;
	for(let spell of SortedSpellList[elements_type[selected][2]]){
		let magic = getMagicData()[spell];
		let name = magic.name;
		if(list_select != -1 && spell_id == list_select){
			locked_id = ["§e", ""]
		}else{
			locked_id = ["", ""]
		}
		if(GradeToInt[magic.grade] > spell_mastering){
			locked_id = ["§c", "_locked"]
			name = "§k" + magic.name + "§k";
		}
		table.button((GradeToInt[magic.grade] - 1 <= spell_mastering) ? magic.path : "textures/ui/locked_spell", locked_id[0] + "§l" + name + "\n§r" + locked_id[0] + "[ " + magic.grade.toUpperCase() + " ] amw:spell" + locked_id[1]);
		spell_id++;
	}

	table.body(visibility + " scroll_craft");
	table.show(source).then( r => {
		if(r.selection == undefined) return;
		if(r.selection <= 8){
			scrollCraft(source, r.selection, book);
		}else{
			if(r.selection == 11 && can_craft){
				let item_name = SortedSpellList[elements_type[selected][2]][list_select];
				let item = new ItemStack(ScrollItem[item_name]);
				let container = source.getComponent("minecraft:inventory").container;
				if(container.emptySlotsCount > 0){
					container.addItem(item);
				}else{
					source.dimension.spawnItem(item, source.location);
				}
				if(source.getGameMode() != "creative"){
					let register_item = [];
					let item_temp = JSON.parse(JSON.stringify(spell_recipes));
					for(let i = 0; i < 36; i++){
						let item = container.getSlot(i);
						if(item.hasItem() && item_identifier.includes(item.typeId) && !register_item.includes(item.typeId)){
							let amount = item_temp[item_identifier_index[item.typeId]].amount;
							if(amount <= 0){
								register_item.push(item.typeId);
								continue;
							}

							item_temp[item_identifier_index[item.typeId]].amount -= item.amount;
							if(item.amount <= amount){
								item.setItem();
							}else{
								item.amount -= amount;
							}
						}
					}
				}
				
				scrollCraft(source, selected, book);
			}else{
				scrollCraft(source, selected, book, r.selection - 9 - spell_offset);
			}
		}
	});
}

export function openMagicLearningStation(source, block){
	let dimension = source.dimension;
	let book = 0;
	if(source.getGameMode() == "creative"){
		book = 4;
	}else{
		for(let x = -2; x <= 2; x++){
			if(book >= 25) break;
			for(let y = -2; y <= 2; y++){
				if(book >= 25) break;
				for(let z = -2; z <= 2; z++){
					if(book >= 25) break;
					let scan = dimension.getBlock(Vector.add(block.location, { x: x, y: y, z: z }));
					if(scan.typeId == "minecraft:bookshelf") book++;
				}
			}
		}
		book = Math.min(4, (book / 25) * 4);
	}

	scrollCraft(source, 0, book);
}


system.runInterval(()=>{
    for(let dimension of Dimensions){
        for(let entity of dimension.getEntities({ type: "amw:floating_magic_book" })){
            let location = entity.location;
			try{
                if(dimension.getBlock(location).typeId != "amw:magic_learning_station"){
                    entity.remove();
                    continue;
                }
            }catch(error){
                continue;
            }
		}
	}
}, 2);