import { system } from "@minecraft/server";
import { getMagicData } from "./magic_list";

const Elements = [
    "nature",
    "fire", 
    "void", 
    "dark",
    "water",
    "thunder",
    "wind",
    "light",
    "ice"
];
const ElementToItem = {
    "nature": "amw:nature_magic_essence",
    "fire": "amw:fire_magic_essence", 
    "void": "amw:void_magic_essence", 
    "dark": "amw:dark_magic_essence",
    "water": "amw:water_magic_essence",
    "thunder": "amw:thunder_magic_essence",
    "wind": "amw:wind_magic_essence",
    "light": "amw:light_magic_essence",
    "ice": "amw:ice_magic_essence"
};

const GradeToInt = {
	basic: 0,
	intermediate: 1,
	advance: 2,
	expert: 3,
	master: 4
}

var spell_recipes = {};

export function updateSpellRecipes(){
    let magic_data;
    try{
        magic_data = getMagicData();
    }catch(err){
		system.runTimeout(()=>{
			updateSpellRecipes()
		}, 2);
        return;
    }
    let spell_id = Object.keys(magic_data);
    for(let spell_name of spell_id){
        if(spell_name == "none") continue;
        let spell = magic_data[spell_name];
        let data = [];

        data.push({
            identifier: "minecraft:paper",
            amount: Math.ceil(spell.cast_duration / 10)
        });
        let essence = ((spell.soul_cost + spell.cast_duration + GradeToInt[spell.grade] * 100) / 30);
        let used_elements = [];
        let total_elements = 0;
        for(let attribute of spell.attribute){
            if(Elements.includes(attribute)){
                used_elements.push(attribute);
                total_elements++;
            }
        }
        let total = Math.ceil(essence / total_elements);
        for(let attribute of used_elements){
            data.push({
                identifier: ElementToItem[attribute],
                amount: total
            });
        }
        // let data = {
        //     paper: paper,
        //     elements: used_elements,
        //     essence: Math.ceil(essence / total_elements)
        // }
        spell_recipes[spell_name] = data;
    }
}

updateSpellRecipes();

export function getSpellRecipes(){
    return spell_recipes;
}